/*
 * A file accessor employing a file provided by an Ext2-compatible filesystem.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <string.h>

#include <algorithm>

#include <systypes/fcntl.h>

#include "ext2_file_accessor.h"



Ext2FileAccessor::Ext2FileAccessor(Ext2FileOperations *ops, ext2_file_t file,
                                   fileid_t fileid)
: Accessor(fileid), _ops(ops), _file(file)
{
}

/* Return the size of the file. */

offset_t Ext2FileAccessor::get_size()
{
    return _ops->get_size(_file);
}

/* Update the size of the file. */

void Ext2FileAccessor::set_size(offset_t size)
{
    _ops->set_size(_file, size);
}

/* Perform any closing operation on the file. */

void Ext2FileAccessor::close()
{   
    if (_closed)
        return;

    _ops->close_file(_file);
    _closed = true;
}

/* Data transfer helper methods. */

/* Fill the populated portion of a flexpage. */

void Ext2FileAccessor::fill_populated(Flexpage *flexpage)
{
    offset_t filepos = flexpage->base_offset;
    offset_t addr = flexpage->base_addr;

    /* Tag the region with file state. */

    flexpage->region->fill(fileid, filepos);

    /* Fill the region with file content. */

    offset_t nread = _ops->read_file(_file, filepos, (void *) addr, flexpage->size);

    /* Pad the flexpage with zero. */

    memset((void *) (addr + nread), 0, flexpage->size - nread);
}

/* Flush the populated portion of a flexpage. */

void Ext2FileAccessor::flush_populated(Flexpage *flexpage)
{
    offset_t filepos = flexpage->base_offset;
    offset_t addr = flexpage->base_addr;
    offset_t populated_size = std::min(flexpage->size, get_size() - filepos);

    /* Remove the file state tag from the region. */

    flexpage->region->flush();

    /* Copy the populated region to the file. */

    _ops->write_file(_file, filepos, (const void *) addr, populated_size);
}

// vim: tabstop=4 expandtab shiftwidth=4
